/*
 * setfont.c - Eugene Crosser & Andries Brouwer
 *
 * Version 0.90
 *
 * Loads the console font, and possibly the corresponding screen map(s).
 * We accept two kind of screen maps, one [-m] giving the correspondence
 * between some arbitrary 8-bit character set currently in use and the
 * font positions, and the second [-u] giving the correspondence between
 * font positions and Unicode values.
 *
 * m68k adaptions by Roman Hodek 03/04/96:
 * On m68k machines, there's no EGA/VGA font generator... But the kernel
 * provides an ioctl PIO_XFONT, that can also set the font width/height,
 * activate kernel predefined fonts and copy fonts from other consoles
 * (conserving memory). There were so much changes that I didn't keep the
 * PIO_FONT interface as an alternative.
 */
#include <stdio.h>
#include <memory.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/kd.h>
#include "paths.h"
#include "psf.h"
#include <linux/kd.h>

char *progname;

static void position_codepage(int iunit, FILE *fpi);
static void loadnewfont(char *ifil, int iunit, int no_m, int no_u );
void do_predeffont(char *name);
void do_copyfont(int cnr);
extern void saveoldmap(char *omfil);
extern void loadnewmap(char *mfil);
extern void saveunicodemap(int fd, char *oufil);
extern void loadunicodemap(int fd, char *ufil);
extern void activate_user_map(void);
extern void activate_def_map(void);
extern char *malloc();

/* search for the font in these directories (with trailing /) */
char *fontdirpath[] = { "", DATADIR "/" FONTDIR "/", 0 };
char *fontsuffixes[] = { "", ".psf", ".cp", ".fnt", ".psfu", 0 };

#define MAX_FONT_HEIGHT	64	/* arbitrary */

static inline FILE*
findfont(char *fnam) {
    return findfile(fnam, fontdirpath, fontsuffixes);
}


void
usage(void)
{
        fprintf(stderr, "\
Usage:  %s [-fontsize] [-p] [newfont] [-m consolemap] [-u unicodemap]
If no newfont is given, then the font \"default\" is loaded. Explicitly
(with -m or -u) or implicitly (in the fontfile) given mappings will
be loaded and, in the case of consolemaps, activated.
Options:
 -8, -14, -16	Choose the right font from a codepage that contains three fonts
                (or choose default font, e.g. \"default8x16\").
 -m none	Suppress loading and activation of a mapping table.
 -u none	Suppress loading of a unicode map.
 -p             Engage a kernel predefined font with name <newfont>.
 -c num         Reuse font from console nr. <num>.
", progname);
	exit(1);
}

int
main(int argc, char *argv[])
{
	char *ifil, *mfil, *ufil;
	int i, iunit, no_m, no_u, predef, copy;

	progname = argv[0];

	ifil = mfil = ufil = 0;
	iunit = 0;
	no_m = no_u = predef = 0;
	copy = -1;

	for (i = 1; i < argc; i++) {
	    if (!strcmp(argv[i], "-m")) {
		if (++i == argc || mfil)
		  usage();
		if (!strcmp(argv[i], "none"))
		  no_m = 1;
		else
		  mfil = argv[i];
	    } else if (!strcmp(argv[i], "-u")) {
		if (++i == argc || ufil)
		  usage();
		if (!strcmp(argv[i], "none"))
		  no_u = 1;
		else
		  ufil = argv[i];
	    } else if (!strcmp(argv[i], "-p")) {
		if (predef)
		  usage();
		predef = 1;
	    } else if (!strcmp(argv[i], "-c")) {
		if (++i == argc || copy >= 0)
		  usage();
		copy = atoi(argv[i]);
		if (copy < 0)
		  usage();
	    } else if(argv[i][0] == '-') {
		iunit = atoi(argv[i]+1);
		if(iunit <= 0 || iunit > MAX_FONT_HEIGHT)
		  usage();
	    } else {
		if (ifil)
		  usage();
		ifil = argv[i];
	    }
	}

	if (copy >= 0) {
	    do_copyfont( copy );
	    return( 0 );
	}
	
	if (predef) {
		if (!ifil) {
		    fprintf( stderr, "Font name missing\n" );
		    exit(1);
		}
		do_predeffont( ifil );
		activate_def_map();
		return( 0 );
	}
	
	if (!ifil && !mfil && !ufil)
	  /* reset to some default */
	  ifil = "";

	if (mfil) {
	    loadnewmap(mfil);
	    activate_user_map();
	    no_m = 1;
	}

	if (ufil)
	  no_u = 1;

	if (ifil)
	  loadnewfont(ifil, iunit, no_m, no_u);

	if (ufil)
	  loadunicodemap(0, ufil);

	return 0;
}

void
do_loadfont(int unit, FILE *fpi) {
	struct linux_xconfont *xfont;

	if (unit < 1 || unit > MAX_FONT_HEIGHT) {
	    fprintf(stderr, "Bad character size %d\n", unit);
	    exit(1);
	}

	if (!(xfont = malloc( sizeof(struct linux_xconfont) + 256*unit ))) {
	    fprintf( stderr, "Out of memory\n" );
		exit(1);
	}
	if (fread(xfont->data, unit, 256, fpi) != 256) {
		perror("Cannot read font from file");
		exit(1);
	}

	fprintf(stderr,"Loading 8x%d font from file %s\n", unit, pathname);

	xfont->width = 8;
	xfont->height = unit;
	if (ioctl(0,PIO_XFONT,xfont)) {
		if (errno == ENOENT)
			fprintf( stderr, "PIO_XFONT ioctl error: no such font\n" );
		else if (errno == ENXIO)
			fprintf( stderr, "PIO_XFONT ioctl error: cannot activate font\n" );
		else
			perror("PIO_XFONT ioctl error");
	    exit(1);
	}
}

void
do_predeffont(char *name) {
	struct linux_xconfont *xfont;

	if (!(xfont = malloc( sizeof(struct linux_xconfont) + strlen(name)+1 ))) {
	    fprintf( stderr, "Out of memory\n" );
		exit(1);
	}
	strcpy( xfont->data, name );

	fprintf(stderr,"Engaging predefined font %s\n", name);

	xfont->width = 0;
	xfont->height = 0;
	if (ioctl(0, PIO_XFONT,xfont)) {
		if (errno == ENOENT)
			fprintf( stderr, "PIO_XFONT ioctl error: no such font\n" );
		else if (errno == ENXIO)
			fprintf( stderr, "PIO_XFONT ioctl error: cannot activate font\n" );
		else
			perror("PIO_XFONT ioctl error");
	    exit(1);
	}
}

void
do_copyfont(int cnr) {
	struct linux_xconfont xfont;

	fprintf(stderr,"Copying font from console %d\n", cnr);

	xfont.width = 1;
	xfont.height = cnr;
	if (ioctl(0,PIO_XFONT,&xfont)) {
		if (errno == ENOENT)
			fprintf( stderr, "PIO_XFONT ioctl error: no such font\n" );
		else if (errno == ENXIO)
			fprintf( stderr, "PIO_XFONT ioctl error: cannot activate font\n" );
		else if (errno == ENOTTY)
			fprintf( stderr, "PIO_XFONT ioctl error: no such console\n" );
		else
			perror("PIO_XFONT ioctl error");
	    exit(1);
	}
}

void
do_loadtable(int tailsz, int fontsize, FILE *fpi) {
	struct unimapinit advice;
	struct unimapdesc ud;
	struct unipair *up;
	int ct = 0, maxct;
	int glyph, i;
	u_short unicode;

	maxct = tailsz;		/* more than enough */
	up = (struct unipair *) malloc(maxct * sizeof(struct unipair));
	if (!up) {
	    fprintf(stderr, "Out of memory?\n");
	    exit(1);
	}
	for (glyph = 0; glyph < fontsize; glyph++) {
	    while ((i = fread(&unicode, sizeof(unicode), 1, fpi)) == 1 &&
		   unicode != PSF_SEPARATOR) {
		if (ct >= maxct) { /* impossible */
		    i = 0;
		    break;
		}
		up[ct].unicode = unicode;
		up[ct].fontpos = glyph;
		ct++;
	    }
	    if (i != 1) {
		fprintf(stderr, "Input error reading unicode table\n");
		exit(1);
	    }
	}

	advice.advised_hashsize = 0;
	advice.advised_hashstep = 0;
	advice.advised_hashlevel = 0;
	if(ioctl(0, PIO_UNIMAPCLR, &advice)) {
	    if (errno == ENOIOCTLCMD) {
		fprintf(stderr, "It seems this kernel is older than 1.1.92\n");
		fprintf(stderr, "No Unicode mapping table loaded.\n");
	    } else
	      perror("PIO_UNIMAPCLR");
	    exit(1);
	}
	fprintf(stderr,"Loading Unicode mapping table...\n");
	ud.entry_ct = ct;
	ud.entries = up;
	if(ioctl(0, PIO_UNIMAP, &ud)) {
#if 0
	    if (errno == ENOMEM) {
		/* change advice parameters */
	    }
#endif
	    perror("PIO_UNIMAP");
	    exit(1);
	}
}

void
loadnewfont(char *ifil, int iunit, int no_m, int no_u) {
	FILE *fpi;
	char defname[20];
	int size, unit;
	struct stat stbuf;

	if (!*ifil) {
	    /* try to find some default file */

	    if (iunit < 0 || iunit > MAX_FONT_HEIGHT)
	      iunit = 0;
	    if (iunit == 0) {
		if ((fpi = findfont(ifil = "default")) == NULL &&
		    (fpi = findfont(ifil = "default8x16")) == NULL &&
		    (fpi = findfont(ifil = "default8x14")) == NULL &&
		    (fpi = findfont(ifil = "default8x8")) == NULL) {
		    fprintf(stderr, "Cannot find default font\n");
		    exit(1);
		}
	    } else {
		sprintf(defname, "default8x%d", iunit);
		if ((fpi = findfont(ifil = defname)) == NULL) {
		    fprintf(stderr, "Cannot find %s font\n", ifil);
		    exit(1);
		}
	    }
	} else {
	    if ((fpi = findfont(ifil)) == NULL) {
		fprintf(stderr, "Cannot open font file %s\n", ifil);
		exit(1);
	    }
	}

	if (stat(pathname, &stbuf)) {
	    perror(pathname);
	    exit(1);
	}
	size = stbuf.st_size;

	/* test for psf first */
	{
	    struct psf_header psfhdr;
	    int fontsize;
	    int hastable;
	    int head;

	    if (fread(&psfhdr, sizeof(struct psf_header), 1, fpi) != 1) {
		perror("Error reading header input font");
		exit(1);
	    }
#ifdef __mc68000__
	    if (psfhdr.magic != PSF_MAGIC_BE) {
#else
	    if (psfhdr.magic != PSF_MAGIC) {
#endif
		fseek(fpi,0,SEEK_SET);
		goto no_psf;
	    }
	    if (psfhdr.mode > PSF_MAXMODE) {
		fprintf(stderr, "Unsupported psf file mode\n");
		exit(1);
	    }
	    fontsize = ((psfhdr.mode & PSF_MODE512) ? 512 : 256);
	    if (fontsize != 256) {
		fprintf(stderr, "Only fontsize 256 supported\n");
		exit(1);
	    }
	    hastable = (psfhdr.mode & PSF_MODEHASTAB);
	    unit = psfhdr.charsize;
	    head = sizeof(struct psf_header) + fontsize*unit;
	    if (head > size || (!hastable && head != size)) {
		fprintf(stderr, "Input file: bad length\n");
		exit(1);
	    }
	    do_loadfont(unit, fpi);
	    if (hastable && !no_u)
	      do_loadtable(size-head, fontsize, fpi);
	    fclose(fpi);
	    return;
	}
      no_psf:

	/* file with three code pages? */
	if (size == 9780) {
	    position_codepage(iunit, fpi);
	    unit = iunit;
	} else {
	    /* bare font */
	    if (size & 0377) {
		fprintf(stderr, "Bad input file size\n");
		exit(1);
	    }
	    unit = size/256;
	}
	do_loadfont(unit, fpi);
	fclose(fpi);
}

static void
position_codepage(int iunit, FILE *fpi) {
        int offset;

	/* code page: first 40 bytes, then 8x16 font,
	   then 6 bytes, then 8x14 font,
	   then 6 bytes, then 8x8 font */

	if (!iunit) {
	    fprintf(stderr, "\
This file contains 3 fonts: 8x8, 8x14 and 8x16. Please indicate
using an option -8 or -14 or -16 which one you want loaded.\n");
	    exit(1);
	}
	switch (iunit) {
	  case 8:
	    offset = 7732; break;
	  case 14:
	    offset = 4142; break;
	  case 16:
	    offset = 40; break;
	  default:
	    fprintf(stderr, "\
You asked for font size %d, but only 8, 14, 16 are possible here.\n",
		    iunit);
	    exit(1);
	}
	if(fseek(fpi,offset,SEEK_SET)) {
	    perror("seek error on input file");
	    exit(1);
	}
}


/* Only on the current console? On all allocated consoles? */
/* A newly allocated console has NORM_MAP by default -
   probably it should copy the default from the current console?
   But what if we want a new one because the current one is messed up? */
/* For the moment: only the current console, only the G0 set */

/* ++roman: It's a current design deficieny that there's only one user
 * translation... Should be fixed some time in the kernel.
 * Also, set G0 _and_ G1 to avoid ^N trashing the mapping...
 */

void
activate_user_map(void) {
    printf("\033(K\033)K");
}

/* (Re-)activate the standard mappings: NORM_MAP in G0, GRAF_MAP in G1 */

void
activate_def_map(void) {
    printf("\033(B\033)0");
}
